/* -*- mode: c++; c-basic-offset: 4; -*- */
#define RAW_SOCKET 1

#include "sends_main.hh"
#include "dbase.hh"
#include "ParseLine.hh"
#include "file_driver.hh"
#include "fonl_driver.hh"
#include "driver_table.hh"
#include "auth_server.hh"
#include "src_update.hh"
#include "update_list.hh"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <iostream>
#include <cstdlib>
#include <stdexcept>
#ifdef WITH_LSMP
#include "lsmp_driver.hh"
#endif


#ifdef RAW_SOCKET
#include "raw_socket.hh"
#else
#include "sslbase.hh"
#endif

using namespace std;

#define PORT "31200"

//======================================  Sends main function
int
main(int argc, const char* argv[]) {

    //----------------------------------  Parse the demonic arguments
    char const* * argvec = new const char*[argc];
    int nvec = 1;
    argvec[0] = argv[0];
    bool demonize = false;
    const char* logfile = 0;
    for (int i=1; i<argc; ++i) {
	if (!strcmp(argv[i], "-d")) {
	    demonize = true;
	} else if (!strcmp(argv[i], "-l")) {
	    logfile = argv[++i];
	} else {
	    argvec[nvec++] = argv[i];
	}
    }

    //-----------------------------------  Fork a daemon
    if (demonize) {
	switch (fork()) {
	case -1:
	    exit(1);
	case 0:
	    break;
	default:
	    exit(0);
	}
    }

    //-----------------------------------  Optionally switch to a log file
    if (logfile) {
	int flags = O_WRONLY | O_CREAT;
	int fd = open(logfile, flags, 0664);
	close(1); // stdout
	if (dup(fd) != 1) {
	    cerr << "stdout not diverted!" << endl;
	}
	close(2); // stderr
	if (dup(fd) != 2) {
	    cout << "stderr not diverted!" << endl;
	}
	close(fd);
    }

    //-----------------------------------  Fire up the server
    sends::sends_main mClass(nvec, argvec);
    delete[] argvec;
    if (mClass.init_complete()) {
	mClass.doit();
    }
}

bool term=false;

extern "C" {
    void
    ctrlc_int(int sig) {
	term = true;
    }
}

namespace sends {

    const int sends_default_threads = 8;
    const int sends_default_inputQ  = 8;
    const seda_stage::dbg_level sends_default_debug = seda_stage::dlvl_none;
    const double sends_default_timeout = 300.0;
    const int sends_trans_count     = 128;

    //==================================  Sends_main constructor
    sends_main::sends_main(int argc, const char* argv[]) 
      : mConfigure("sends.conf"), mInitComplete(false),
        mAuthPool  (sends_default_threads, sends_default_inputQ),
        mConcPool (sends_default_threads,  sends_default_inputQ),
        mParsePool (sends_default_threads, sends_default_inputQ),
        mLookupPool(sends_default_threads, sends_default_inputQ),
        mFetchPool (sends_default_threads, sends_default_inputQ),
        mWritePool (sends_default_threads, sends_default_inputQ),
	mSmallBufSize(sends_small_buffer_size), 
	mSmallBufCount(sends_small_pool_size),
	mLargeBufSize(sends_large_buffer_size), 
	mLargeBufCount(sends_large_pool_size),
	mTransCount(sends_trans_count),
	mListenAddr(PORT), mListenQueue(8), mAuthType(auth_server::kSasl)
    {
	if (argc > 1) mConfigure = argv[argc-1];

	//----------------------------------  Set a default server name
	auth_server::refGlobal().setServer("nds2");
	auth_server::refGlobal().setRealm("LIGO.ORG");

	//----------------------------------  Build the driver table.
	TheDriverTable.add(new sends::file_driver);

	//----------------------------------  Set stage maps
	seda_stage::stage_pars def_pars;
	def_pars._debug        = sends_default_debug;
	def_pars._inq_capacity = sends_default_inputQ;
	def_pars._threads      = sends_default_threads;
	def_pars._timeout      = sends_default_timeout;

	mStageParMap.insert(stage_par_node("authorize",    def_pars));
	mStageParMap.insert(stage_par_node("concentrator", def_pars));
	mStageParMap.insert(stage_par_node("parser",       def_pars));
	mStageParMap.insert(stage_par_node("lookup",       def_pars));
	mStageParMap.insert(stage_par_node("fetch",        def_pars));
	mStageParMap.insert(stage_par_node("write",        def_pars));

	//----------------------------------  Read the configuration file.
	cout << endl << "Configuring from file: " << mConfigure << endl;
	try {
	    configure(mConfigure);
	}
	catch (exception& e) {
	    cerr << "Caught exception during configuration: " 
		 << e.what() << endl;
	    return;
	}

	//----------------------------------  Build channel cache pools
	Buffer_Pool.add_pool(seg_pool_mgr::kSmallBuf, mSmallBufCount, 
			     mSmallBufSize);
	Buffer_Pool.add_pool(seg_pool_mgr::kLargeBuf, mLargeBufCount, 
			     mLargeBufSize);

	//----------------------------------  Set channel list size for seg_db
	Segment_List.set_channel_count(Channel_List.size());

	//----------------------------------  Set stage parameters
	mAuthPool.set_parameters(mStageParMap["authorize"]);
	mConcPool.set_parameters(mStageParMap["concentrator"]);
	mParsePool.set_parameters(mStageParMap["parser"]);
	mLookupPool.set_parameters(mStageParMap["lookup"]);
	mFetchPool.set_parameters(mStageParMap["fetch"]);
	mWritePool.set_parameters(mStageParMap["write"]);

	//----------------------------------  Link the pipeline stages
	mAuthPool.link_output(mConcPool);
	mConcPool.link_output(mParsePool);
	mParsePool.link_output(mLookupPool);
	mParsePool.link_alternate(mConcPool);
	mLookupPool.link_output(mFetchPool);
	mFetchPool.link_output(mWritePool);
	mWritePool.link_output(mConcPool);
	mWritePool.link_alternate(mFetchPool);
	mParsePool.link_error(mConcPool);
	mLookupPool.link_error(mConcPool);
	mFetchPool.link_error(mConcPool);
	mWritePool.link_error(mConcPool);

	//----------------------------------  Make a free-transaction pool
	tranQ.resize(mTransCount);
	tranVec.reserve(mTransCount);
	for (int i=0; i<mTransCount; ++i) {
	    transaction* t = new sends::transaction(tranQ, mAuthType);
	    tranQ.push(t);
	    tranVec.push_back(t);
	}

	//----------------------------------  Dump memory usage
	cout << "Memory usage at start: " << endl;
	cout << "  Channel list:        " << Channel_List.mem_use() << endl;
	cout << "  Source list:         " << Source_List.mem_use() << endl;

	//----------------------------------  Start all the threads
	sigignore(SIGPIPE);
	signal(SIGINT,  ctrlc_int);
	signal(SIGTERM, ctrlc_int);
#ifdef linux
	if (siginterrupt(SIGINT, 1))  perror("siginterrupt(SIGINT, on)");
	if (siginterrupt(SIGTERM, 1)) perror("siginterrupt(SIGTERM, on)");
	//if (siginterrupt(SIGPIPE, 1)) perror("siginterrupt(SIGPIPE, on)");
#endif

	//----------------------------------  Start data update
	UpdateList.start_all();

	//----------------------------------  Start all the threads
	mAuthPool.start();
	mParsePool.start();
	mLookupPool.start();
	mFetchPool.start();
	mWritePool.start();
	mInitComplete = true;
    }

    //==================================  Sends_main destructor
    sends_main::~sends_main(void) 
    {
	//------------------------------------  Shut down all stages
	mAuthPool.shut();
	mConcPool.shut();
	mParsePool.shut();
	mLookupPool.shut();
	mFetchPool.shut();
	mWritePool.shut();

	//----------------------------------  Start data update
	UpdateList.stop_all();

	//------------------------------------  Free the transaction queue
	sends::transaction* trans = 0;
	long nDelete = 0;
	while (tranQ.pop_if_present(trans)) {
	    delete trans;
	    for (unsigned int j=0; j<tranVec.size(); ++j) {
		if (tranVec[j] == trans) tranVec[j] = 0;
	    }
	    nDelete++;
	}

	if (nDelete != mTransCount) {
	    cout << "Waiting for lagging transactions ...";
	    sleep(22);
	    while (tranQ.pop_if_present(trans)) {
		delete trans;
		for (unsigned int j=0; j<tranVec.size(); ++j) {
		    if (tranVec[j] == trans) tranVec[j] = 0;
		}
		nDelete++;
	    }
	    cout << endl;
	}

	if (nDelete != mTransCount) {
	    cout << "Undeleted transactions: " << endl;
	    for (unsigned int j=0; j<tranVec.size(); ++j) {
		trans = tranVec[j];
		if (trans) {
		    cout << "id: " << trans->ident() << " last stage: " 
			 << trans->get_stage() << endl;
		}
	    }
	}

	//------------------------------------  Clear the segment list
	//Segment_List.clear();
    }

    //==================================  Pick an entry in a 
    string
    pick(const string& in, int N) {
	string::size_type idot = 0;
	for (int i=0; i<N; ++i) {
	    idot = in.find('.', idot);
	    if (idot != string::npos) idot++;
	}
	if (idot == string::npos) return string();
	string::size_type iend = in.find('.', idot) - idot;
	return in.substr(idot, iend);
    }

    //==================================  Configure some stuff
    void 
    sends_main::configure(const std::string& file) {
	typedef std::pair<std::string, long> prot_node;
	std::vector<prot_node> protect_list;

	static int depth(0);
	ParseLine pl(file.c_str());
	if (!pl.isOpen()) {
	    cerr << "Unable to open file: " << mConfigure << endl;
	    return;
	}

	//------------------------------  Loop over config file entries
	while (pl.getLine() >= 0) {
	    int nArg = pl.getCount();
	    if (nArg < 2) continue;
	    string lvl0 = pick(pl[0], 0);
	    string lvl1 = pick(pl[0], 1);
	    string lvl2 = pick(pl[0], 2);

	    //--------------------------  Buffer class data
	    if (lvl0 == "buffer") {
		if (lvl1 == "small") {
		    if (lvl2 == "count") {
			mSmallBufCount = pl.getInt(1);
		    } else if (lvl2 == "size") {
			mSmallBufSize = pl.getInt(1);
		    } else {
			cerr << "Unrecognized buffer parameter: " 
			     << lvl2 << endl;
		    }
		} else if (lvl1 == "large") {
		    if (lvl2 == "count") {
			mLargeBufCount = pl.getInt(1);
		    } else if (lvl2 == "size") {
			mLargeBufSize = pl.getInt(1);
		    } else {
			cerr << "Unrecognized buffer parameter: " 
			     << lvl2 << endl;
		    }
		} else {
		    cerr << "Unrecognized buffer class: " << lvl1 << endl;
		}

	    //--------------------------  Get channel class data
	    } else if (lvl0 == "channel") {
		if (lvl1 == "access") {
		    prot_node pchan(pl[1], pl.getInt(2));
		    protect_list.push_back(pchan);
		} else if (lvl1 == "file") {
		    epoch ep;
		    if (nArg == 3) {
			ep = epoch(pl.getInt(2));
		    } else if (nArg > 3) {
			if (string(pl[3]) == "inf" || *pl[3] == '-') {
			    ep = epoch(pl.getInt(2));
			} else {
			    ep = epoch(pl.getInt(2), pl.getInt(3));
			}
		    }
		    Channel_List.read(pl[1], ep);
		} else if (lvl1 == "dump") {
		    if (pl.getInt(1)) Channel_List.dump(std::cerr);
		} else {
		    cerr << "Unrecognized channel parameter: " << lvl1 << endl;
		}
	    }

	    //--------------------------  Manage epoch definitions
	    else if (lvl0 == "epoch") {
		//----------------------  Define an epoch
		if (lvl1 == "define") {
		    epoch epic;
		    if (nArg == 3) {
			epic = epoch::parse(pl[2]);
		    } else if (nArg == 4) {
			epic = epoch(pl.getInt(2), pl.getInt(3));
		    } else {
			cerr << "Too many arguments to epoch_define" << endl;
		    }
		    Epoch_Table.add(pl[1], epic);
		} 

		//----------------------  Print it all out
		else if (lvl1 == "display") {
		    if (pl.getInt(1)) Epoch_Table.display(cout);
		}

		//----------------------  Unrecognized parameter
		else {
		    cerr << "Unrecognized epoch parameter: " << lvl1 << endl;
		}
	    }

	    //--------------------------  Include a file
	    else if (lvl0 == "include") {
		string file = pl[1];
		depth++;
		cout << endl << "Including file: " << file << " (depth " 
		     << depth << ")" << endl;
		configure(file);
		depth--;

	    //--------------------------  Get source class data
	    } else if (lvl0 == "source") {
		if (lvl1 == "driver") {
		    if (lvl2 == "fonl") {
			if (nArg < 3) {
			    cerr <<"Insufficient fonl driver arguments"<< endl; 
			    cerr <<"Syntax: source.driver.fonl <name> <update>" 
				 << endl;
			} else {
			    fonl_driver* dp = new fonl_driver(pl[1]);
			    dp->set_update(pl[2]);
			    TheDriverTable.add(dp);
			}
		    } else if (lvl2 == "lsmp") {
#ifdef WITH_LSMP
			lsmp_driver* dp = new lsmp_driver(pl[1]);
			if (nArg > 2) dp->setDebug(pl.getInt(2));
			TheDriverTable.add(dp);
#else
			cerr << "###################################" << endl;
			cerr << "#                                 #" << endl;
			cerr << "#    lsmp driver not installed!   #" << endl;
			cerr << "#                                 #" << endl;
			cerr << "###################################" << endl;
#endif
		    } else {
			cerr << "Unrecognized driver type: " << lvl2 << endl;
		    }
		} else if (lvl1 == "file") {
		    Source_List.read(pl[1]);
		} else if (lvl1 == "update") {

		    //------------------  Parse source string <arch>:<pfx>:<gps>
		    string fsname = pl[1];
		    loc_list::arch_type a = loc_list::kOnline;
		    gps_type gps = 0;
		    string::size_type inx = fsname.find(":");
		    if (inx != string::npos) {
			if (inx != 0) {
			    a = loc_list::get_arch_type(fsname.substr(0, inx));
			}
			fsname.erase(0,inx+1);
			inx = fsname.find(":");
			if (inx != string::npos) {
			    gps = strtol(fsname.c_str()+inx+1, 0, 0);
			    fsname.erase(inx);
			}
		    }
		    src_update su;
		    su.set_source(a, fsname, gps);

		    //------------------  Parse scheduling parameters
		    double step(0), late(0);
		    if (lvl2.empty() || lvl2 == "scan") {
			if (nArg > 2) late = pl.getDouble(2);
			if (nArg > 3) step = pl.getDouble(3);
			if (nArg > 4) su.set_debug(pl.getInt(4));
			su.set_scan_mode(late, step);
		    } else if (lvl2.empty() || lvl2 == "sched") {
			if (nArg > 2) late = pl.getDouble(2);
			if (nArg > 3) step = pl.getDouble(3);
			if (nArg > 4) su.set_debug(pl.getInt(4));
			su.set_sched_mode(late, step);
		    } else if (lvl2 == "synch") {
			if (nArg < 3) {
			    cout << "No file argument for source.update.synch"
				 << endl;
			    continue;
			}
			if (nArg > 3) late = pl.getDouble(3);
			if (nArg > 4) step = pl.getDouble(4);
			if (nArg > 5) su.set_debug(pl.getInt(5));
			su.set_synch_mode(pl[2], late,step);
		    } else {
			cout << "Invalid update scheduler type: " << lvl2
			     << endl;
			continue;
		    }
		    UpdateList.add(su);
		} else if (lvl1 == "dump") {
		    int lvl = pl.getInt(1);
		    if (lvl) Source_List.dump(std::cerr, lvl);
		} else {
		    cerr << "Unrecognized channel parameter: " << lvl1 << endl;
		}

	    //--------------------------  Get server class data
	    } else if (lvl0 == "server") {
		if (lvl1 == "authentication") {
		    if (lvl2 == "name") {
			auth_server::refGlobal().setServer(pl[1]);
		    } else if (lvl2 == "fqdn") {
			auth_server::refGlobal().setFQDN(pl[1]);
		    } else if (lvl2 == "keytab") {
			setenv("KRB5_KTNAME", pl[1], 1);
		    } else if (lvl2 == "krb5conf") {
			setenv("KRB5_CONFIG", pl[1], 1);
		    } else if (lvl2 == "mechanism") {
			auth_server::refGlobal().addProtocol(pl[1]);
		    } else if (lvl2 == "realm") {
			auth_server::refGlobal().setRealm(pl[1]);
		    } else if (lvl2 == "type") {
			string arg1 = pl[1];
			if (arg1 == "none") mAuthType = auth_server::kNone;
		    } else if (lvl2 == "userfile") {
			User_List.readFile(pl[1]);
		    } else {
			cerr << "Invalid server.authentication subparameter: " 
			     << lvl2 << endl;
		    }
		} else if (lvl1 == "listen") {
		    if (lvl2 == "address") {
			mListenAddr = pl[1];
		    } else if (lvl2 == "allow") {
			for (int i=1; i<nArg; ++i) {
			    unsigned int ipaddr   = 0;
			    struct in_addr netmask;
			    unsigned short port = 0;
			    string addrmask = pl[i];
			    string::size_type inx = addrmask.find("/");
			    if (inx == string::npos) {
				socket_api::parse_addr(addrmask, ipaddr, port);
				netmask.s_addr = ~0;
			    } else {
				socket_api::parse_addr(addrmask.substr(0, inx),
						       ipaddr, port);
				inet_aton(addrmask.substr(inx+1).c_str(), 
					  &netmask);
			    }
			    allow_node anode(ipaddr, netmask.s_addr);
			    mAllowVect.push_back(anode);
			}
		    } else if (lvl2 == "dump") {
			cout << "Allow list:";
			int N=mAllowVect.size();
			if (!N) cout << " ALL";
			for (int i=0; i<N; ++i) {
			    in_addr addr, mask;
			    addr.s_addr = mAllowVect[i].first;
			    mask.s_addr = mAllowVect[i].second;
			    cout << " " << inet_ntoa(addr);
			    cout << "/" << inet_ntoa(mask);
			}
			cout << endl;
		    } else if (lvl2 == "queue") {
			mListenQueue = pl.getInt(1);
		    } else {
			cerr << "Invalid server.listen sub-parameter: " 
			     << lvl2 << endl;
		    }
		} else {
		    cerr << "Unrecognized server parameter: " << lvl1 << endl;
		}

	    //--------------------------  Get stage class data
	    } else if (lvl0 == "stage") {
		stage_par_iter p = mStageParMap.find(lvl1);
		if (p == mStageParMap.end()) {
		    cerr << "Stage: " << lvl1 << " unknown." << endl;
		} else if (lvl2 == "debug") {
		    seda_stage::dbg_level dl = seda_stage::dlvl_none;
		    if (*pl[1] >= '0' && *pl[1] <= '9') {
			dl = static_cast<seda_stage::dbg_level>(pl.getInt(1));
		    } else {
			dl = seda_stage::dbg_lvl_code(pl[1]);
		    }
		    p->second._debug = dl;
		} else if (lvl2 == "q-capacity") {
		    p->second._inq_capacity = pl.getInt(1);
		} else if (lvl2 == "threads") {
		    p->second._threads = pl.getInt(1);
		} else if (lvl2 == "timeout") {
		    p->second._timeout = pl.getDouble(1);
		} else {
		    cerr << "Unrecognized stage parameter: " << lvl2 << endl;
		}

	    //--------------------------  Unknown class
	    } else {
		cerr << "Unknown configure class: " << lvl0 << endl;
	    }
	}

	//------------------------------  Mark protected channels
	index_type N=protect_list.size();
	for (index_type i=0; i<N; i++) {
	    string chname = protect_list[i].first;
	    long   chacc  = protect_list[i].second;
	    chan_index lb = Channel_List.lower_bound(chname);
	    chan_index ub = Channel_List.upper_bound(chname + "z");
	    for (chan_index j=lb; j<ub; j++) {
		cout << "Protect channel: " << Channel_List[j]
		     << " access= " << chacc << endl;
		const_cast<chan_entry&>(Channel_List[j]).set_access(chacc);
	    }
	}
    }

    //==================================  Do it all
    void 
    sends_main::doit(void) 
    {
	//------------------------------  Catch any (fatal) errors
	sends::transaction* trans = 0;
	try {

#ifdef RAW_SOCKET
	    //--------------------------  Open a raw socket and listen to it
	    sends::raw_socket bio;
	    if (bio.listen(mListenAddr, mListenQueue)) {
		perror("Error in listen");
		cerr << "Listen to port: " << mListenAddr << " failed. " 
		     << endl;
		throw runtime_error("Failure to listen to socket");
	    }
#else
	    //--------------------------  SSL initialization: 
	    //                            Set up the ssl context
	    sends::ssl_ctx ssl;
	    ssl.secure("/home/jzweizig/.ssl/certs/ca.cer", 
		       "/home/jzweizig/.ssl/keys/ca.key", 
		       "sends master");

	    //--------------------------  Open an ssl port and listen to it
	    sends::ssl_bio bio(ssl);
	    bio.listen(mListenAddr, mListenQueue);
#endif
	    cout << "Now listening on port: " << mListenAddr << endl;

	    //--------------------------  Loop over connections.
	    long tranID = 0;
	    while( !term ) {
		//----------------------  Accept a socket connection
		trans = 0;
		sends::socket_api* acc = bio.accept();
		if (acc) {

		    //------------------  Check if allowed 
		    bool allow = false;
		    int N = mAllowVect.size();
		    if (!N) {
			allow = true;
		    } else {
			unsigned int addr;
			unsigned short port;
			socket_api::parse_addr(acc->peer_name(), addr, port);
			for (int i=0; i<N; ++i) {
			    unsigned int ipad = mAllowVect[i].first;
			    unsigned int mask = mAllowVect[i].second;
			    if ((addr & mask) == ipad) {
				allow = true;
				break;
			    }
			}
		    }
		    if (!allow) {
			cerr << "connection from: " << acc->peer_name()
			     << " is not allowed!" << endl;
			delete acc;
			continue;
		    }

		    //------------------  Get a free transaction. Set ID, BIO
		    tranQ.pop(trans);
		    if (!trans) {
			throw runtime_error("No transaction objects");
		    }
		    trans->set_id(tranID++);
		    trans->set_socket(acc);
		    trans->set_stage(10);

		    //------------------  Pass it to the authentication stage.
		    if (!mAuthPool.ref_input().push_if_not_full(trans)) {
			acc->write("0016", 4);
			cout << "Transaction " << trans->ident() << " from "
			     << acc->peer_name() << " aborted. " 
			     << "Authorization input queue full." << endl;
			trans->set_stage(20);
			trans->release();
		    }
		}
	    }
	    bio.close();

	} catch (std::exception& e) {
	    if (trans) {
		trans->set_stage(30);
		trans->release();
	    }
	    cout << "Sends_main: Exception: " << e.what() << endl;
	}
    }
}
