/* -*- mode: c++; c-basic-offset: 4; -*- */

/**  This program merges channel lists for a given frame type code and in 
 *  a specified directory.
 *
 *  The arguments are:
 *  channel_merge --index-file <ifile> --list-directory <ldir> \
 *                --rebuild --outfile <ofile> <frame-type>
 */
#include <algorithm>
#include <string>
#include <vector>
#include <stdexcept>
#include <stdint.h>
#include <iostream>
#include <sstream>
#include <fstream>
#include "chan_list.hh"
#include "ParseLine.hh"
#include "scandir.hh"
#include "checksum_crc32.hh"

class ChanMerge {
public:
    typedef unsigned long gps_type;
public:
    ChanMerge(int argc, const char* argv[]);
    ~ChanMerge(void);
    void coalesce(void);
    void doit(void);
    void read(void);
    void rebuild(void);
    void sort(void);
    void write_index(const std::string& file);
    void write_index(std::ostream& out);
public:
    class indexEntry {
    public:
	indexEntry(const std::string& name);
	indexEntry(const std::string& dir, const std::string& type, 
		   const sends::epoch& epic);
	~indexEntry(void);
	bool bustup(std::string& dir, std::string& type, std::string& sufx, 
		    sends::epoch& epic) const;
	uint32_t crc32(void) const;
	void calc_crc(void);
	gps_type end_gps(void) const;
	const std::string& file(void) const;
	void merge(const indexEntry& ent);
	bool operator<(const indexEntry& ie) const;

	/**  Test for overlapping or adjacent entries.
	 */
	bool overlap(const indexEntry& ent) const;
	const sends::epoch& refEpoch(void) const;
	void set_crc32(uint32_t crc);
	void set_span(gps_type start, gps_type dt);
	gps_type start_gps(void) const;
    private:
	std::string _fileName;
	sends::epoch _epic;
	uint32_t    _crc32;
    };

    typedef std::vector<indexEntry>    index_vect;
    typedef index_vect::iterator       index_iter;
    typedef index_vect::const_iterator const_index_iter;

private:
    std::string _frameType;
    std::string _indexFile;
    std::string _listDirectory;
    std::string _outputFile;
    bool        _rebuild;
    index_vect  _indexVector;
};

inline uint32_t
ChanMerge::indexEntry::crc32(void) const {
    return _crc32;
}

inline ChanMerge::gps_type 
ChanMerge::indexEntry::end_gps(void) const {
    return _epic.end();
}

inline const std::string&
ChanMerge::indexEntry::file(void) const {
    return _fileName;
}

inline bool 
ChanMerge::indexEntry::operator<(const indexEntry& ie) const {
    return _epic < ie._epic;
}

inline bool 
ChanMerge::indexEntry::overlap(const indexEntry& ie) const {
    return _crc32 == ie._crc32 && _epic.overlap(ie._epic);
}

inline const sends::epoch&
ChanMerge::indexEntry::refEpoch(void) const {
    return _epic;
}

inline ChanMerge::gps_type 
ChanMerge::indexEntry::start_gps(void) const {
    return _epic.start();
}

using namespace std;
using sends::chan_list;
using sends::chan_entry;

//======================================  Main progragram entry
int
main(int argc, const char* argv[]) {
    try {
	ChanMerge cm(argc, argv);
	cm.doit();
    } catch (std::exception& e) {
	cerr << "Caught exception: " << e.what() << endl;
    }
}

//======================================  indexEntry Constructor
ChanMerge::indexEntry::indexEntry(const std::string& name) 
    : _fileName(name), _crc32(0)
{
    string dir, type, sufx;
    bustup(dir, type, sufx, _epic);
}

//======================================  indexEntry Constructor
ChanMerge::indexEntry::indexEntry(const std::string& dir, 
				  const std::string& type, 
				  const sends::epoch& epic) 
{
    ostringstream out;
    out << dir << "/" << type << "-" << epic.start() << "-" 
	<< epic.duration() << ".cls";
    _fileName = out.str();
    _epic = epic;
    _crc32 = 0;
}

//======================================  indexEntry Destructor
ChanMerge::indexEntry::~indexEntry(void) {
}

//======================================  Parse channel list file name
bool
ChanMerge::indexEntry::bustup(string& dir, string& type, string& sufx, 
			      sends::epoch& epic) const {
    bool nameok = true;
    string tempstr = _fileName;

    //----------------------------------  Get the directory name
    string::size_type inx = tempstr.rfind('/');
    if (inx != string::npos) {
	dir = tempstr.substr(0, inx);
	tempstr.erase(0, inx+1);
    }
    else {
	dir.clear();
    }

    //-----------------------------------  Get the file extension
    inx = tempstr.rfind('.');
    if (inx != string::npos) {
	sufx = tempstr.substr(inx+1);
	tempstr.erase(inx);
    }
    else {
	sufx.clear();
	nameok = false;
    }

    //-----------------------------------  Pick out the duration.
    sends::gps_type start=0, dt=0;
    inx = tempstr.rfind('-');
    if (inx != string::npos) {
	char* estr = 0;
	dt = strtoul(tempstr.c_str()+inx+1, &estr, 10);
	if (!dt || *estr != 0) {
	    nameok = false;
	}
	tempstr.erase(inx);
    }
    else {
	nameok = false;
    }

    //-----------------------------------  Pick out the start gps
    inx = tempstr.rfind('-');
    if (inx != string::npos) {
	char* estr = 0;
	start = strtoul(tempstr.c_str()+inx+1, &estr, 10);
	if (!dt || *estr != 0) {
	    nameok = false;
	}
	tempstr.erase(inx);
    }
    else {
	nameok = false;
    }
    epic = sends::epoch(start, start+dt);

    //-----------------------------------  The rest is a type code
    type = tempstr;

    return nameok;
}

//======================================  Set index gps times.
void 
ChanMerge::indexEntry::calc_crc(void) {
    ifstream in(file().c_str());
    checksum_crc32 crc;
    while (!in.eof()) {
	string line;
	getline(in, line);
	line += '\n';
	crc.add(line);
    }
    _crc32 = crc.result();
}

//======================================  Set index gps times.
void 
ChanMerge::indexEntry::merge(const indexEntry& ent) {
    if (_epic.overlap(ent._epic) && _epic.duration() < ent._epic.duration()) {
	_fileName = ent._fileName;
    }
    _epic.merge(ent._epic);
}

//======================================  Set index gps times.
void
ChanMerge::indexEntry::set_crc32(uint32_t crc) {
    _crc32 = crc;
}

//======================================  Set index gps times.
void
ChanMerge::indexEntry::set_span(gps_type start, gps_type end) {
    _epic = sends::epoch(start, end);
}

//======================================  Constructor
ChanMerge::ChanMerge(int argc, const char* argv[]) {
    int j=1;
    bool syntax = false;
    for (int i=1; i<argc && *argv[i] == '-'; i++) {
	string argi = argv[i];
	if (argi == "--index-file") {
	    if (++i < argc) _indexFile = argv[i];
	    else syntax = true;
	}
	else if (argi == "--list-directory") {
	    if (++i < argc) _listDirectory = argv[i];
	    else syntax = true;
	}
	else if (argi == "--rebuild") {
	    _rebuild = true;
	}
	else if (argi == "--outfile") {
	    if (++i < argc) _outputFile = argv[i];
	    else syntax = true;
	}
	else {
	    syntax = true;
	    break;
	}
	j = i+1;
    }

    if (!syntax && j<argc) {
	_frameType = argv[j];
    } else {
	cerr << "Error in command line syntax: " << endl;
	cerr << "channel_merge --index-file <ifile> --list-directory <ldir> \\"
	     << endl;
	cerr << "              --rebuild --outfile <ofile> <frame-type>" 
	     << endl;
	throw runtime_error("syntax error in command string");
    }


    if (_indexFile.empty()) {
	ostringstream listdir;
	listdir << _listDirectory << "/" << _frameType 
		<< "-IndexFile.txt";
	_indexFile = listdir.str();
    }
}

//======================================  Destructor
ChanMerge::~ChanMerge(void) {
}

//======================================  Coalesce the index.
void
ChanMerge::coalesce(void) {
    sort();
    size_t N =_indexVector.size();
    size_t j = 0;
    for (size_t i=1; i<N; i++) {
	if (_indexVector[j].overlap(_indexVector[i])) {
	    _indexVector[j].merge(_indexVector[i]);
	} else {
	    j++;
	    if (j < i)  _indexVector[j] = _indexVector[i];
	}
    }
    _indexVector.erase(_indexVector.begin()+(++j), _indexVector.end());
}

//======================================  Processing method
void
ChanMerge::doit(void) {
    if (_rebuild) {
	rebuild();
	coalesce();
	write_index(_indexFile);
    }
    else {
	read();
	coalesce();
    }
    if (_indexVector.empty()) {
	cerr << "Empty index list: " << _indexFile << endl;
	return;
    }

   //----------------------------------  Loop over index entries
    chan_list mergelist;
    mergelist.read(_indexVector[0].file(), _indexVector[0].refEpoch());

    size_t N=_indexVector.size();
    for (size_t i=1; i<N; i++) {
	chan_list ci;
	ci.read(_indexVector[i].file(), _indexVector[i].refEpoch());
	chan_list::chan_vect::base_type tba;
	tba.reserve(512);
	size_t minx = 0;
	size_t Nm = mergelist.size();
	size_t Ni = ci.size();
	for (size_t j=0; j<Ni; j++) {
	    while (minx < Nm && mergelist[minx] < ci[j]) minx++;
	    if (minx < Nm && ci[j] == mergelist[minx]) {
		mergelist[minx++].merge(ci[j]);
	    }
	    else {
		tba.push_back(ci[j]);
	    }
	}
	if (!tba.empty()) {
	    mergelist.merge(tba);
	}
    }
    if (_outputFile.empty()) {
	mergelist.dump(cout, true);
    } else {
	ofstream outf(_outputFile.c_str());
	mergelist.dump(outf, true);
	outf.close();
    }         
}

//======================================  Read the index file.
void
ChanMerge::read(void) {
    ParseLine pl(_indexFile.c_str());
    if (!pl.isOpen()) {
	string msg = "Unable to open index file: ";
	msg += _indexFile;
	throw runtime_error(msg);
    }

    while (pl.getLine() >= 0) {
	int nArg = pl.getCount();
	if (!nArg) continue;
	if (nArg < 4) throw runtime_error("Insufficient entries on line");
	indexEntry ie(pl[0]);
	string   dir, type, sufx;
	sends::epoch epic;
	ie.bustup(dir, type, sufx, epic);
	if (type != _frameType || sufx != "cls") {
	    throw runtime_error(string("Invalid file name: ") + pl[0]);
	}
	ie.set_span(pl.getInt(1), pl.getInt(2));
	ie.set_crc32(pl.getInt(3));
	_indexVector.push_back(ie);
    }
}

//======================================  (re)Build an index
void
ChanMerge::rebuild(void) {
    _indexVector.clear();
    _indexVector.reserve(512);
    
    sends::scandir sd(_listDirectory);
    while (sd.next_entry()) {
	//cout << "file path: " << sd.file_path() << endl;
	if (!sd.is_file()) continue;
	
	indexEntry ie(sd.file_path());
	//cout << "entry: " << ie.file() << "  " << ie.start_gps() << "  " 
	//     << ie.end_gps() << endl;
	string dir, type, sufx;
	sends::epoch epic;
	if (!ie.bustup(dir, type, sufx, epic)) continue;
	//cout << "bustup ok" << endl;
	ie.calc_crc();
	_indexVector.push_back(ie);
    }
}

//======================================  Sort the index
void
ChanMerge::sort(void) {
    if (_indexVector.empty()) return;
    std::sort(_indexVector.begin(), _indexVector.end());
}

//======================================  Write the index to a specified file
void 
ChanMerge::write_index(const std::string& file) {
    ofstream of(file.c_str());
    write_index(of);
}

//======================================  Write the index to a standard ostream
void 
ChanMerge::write_index(std::ostream& out) {
    out << "# Channel list index built by channel merge for frame type: "
	<< _frameType << endl;
    size_t N = _indexVector.size();
    for (size_t i=0; i < N; i++) {
	out <<  _indexVector[i].file() << " " <<  _indexVector[i].start_gps() 
	    << " " <<  _indexVector[i].end_gps() << " " 
	    <<  _indexVector[i].crc32() << endl;
    }
}
